// Quad.cpp : implementation of the Quad class
//

#include <string.h>	// memcpy(), memset()

#include "Quad.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// Helper functions

/*
** From Ray Gardner's public domain scalar date functions from the C Snippets
*/
static const LONG SMITHSONIAN_NULL = 2400000L; // == dmy2jdn(17,11,1858)

static inline void jdn2dmy(const LONG jdn, LONG& day, LONG& month, LONG& year)
{
	LONG x, z, m, d, y;
	LONG daysPer400Years = 146097L;
	LONG fudgedDaysPer4000Years = daysPer400Years * 10 + 31;

	x = jdn + 68569L + SMITHSONIAN_NULL + 1;

	z = 4 * x / daysPer400Years;
	x = x - (daysPer400Years * z + 3) / 4;
	y = 4000 * (x + 1) / fudgedDaysPer4000Years;
	x = x - 1461 * y / 4 + 31;
	m = 80 * x / 2447;
	d = x - 2447 * m / 80;
	x = m / 11;
	m = m + 2 - 12 * x;
	y = 100 * (z - 49) + y + x;

	day   = d;
	month = m;
	year  = y;
}

static inline LONG dmy2jdn(const LONG day, const LONG month, const LONG year)
{
	return (LONG)(
		(day - 32076)
		+ 1461L * (year + 4800L + (month - 14) / 12) / 4
		+ 367 * (month - 2 - (month - 14) / 12 * 12) / 12
		- 3 * ((year + 4900L + (month - 14) / 12) / 100) / 4
		- SMITHSONIAN_NULL);
}

/////////////////////////////////////////////////////////////////////////////

static inline ULONG LoWord(const ULONG ul)
	{ return ul & 0xFFFF; }
static inline ULONG HiWord(const ULONG ul)
	{ return ul >> 16; }
static inline ULONG ULongFromWord(const ULONG ulLo, const ULONG ulHi)
	{ return (LoWord(ulLo) | (LoWord(ulHi) << 16)); }

static const int ulong_bits = 8*sizeof(ULONG); // This should better be 32!

static inline void Quad_Shl(const Quad a, const ULONG l, Quad& x)
{
	ULONG ll = l;

	if (ll >= 2*ulong_bits)
	{
		x.lo = x.hi = 0;
		return;
	}

	if (ll >= ulong_bits)
	{
		x.lo = 0;
		if (a.lo == 0)
		{
			x.lo = 0;
			return;
		}
		ll -= ulong_bits;
	}

	x.hi = (a.hi << ll) | (a.lo >> (ulong_bits - ll));
	x.lo = a.lo << ll;
}

static inline void Quad_Shr(const Quad a, const ULONG l, Quad& x)
{
	ULONG ll = l;

	if (ll >= 2*ulong_bits)
	{
		x.lo = x.hi = 0;
		return;
	}

	if (ll >= ulong_bits)
	{
		x.hi = 0;
		if (a.hi == 0)
		{
			x.lo = 0;
			return;
		}
		ll -= ulong_bits;
	}

	x.lo = (a.lo >> ll) | (a.hi << (ulong_bits - ll));
	x.hi = a.hi >> ll;
}

static inline void Quad_Add(const Quad a, const Quad b, Quad& x)
{
	ULONG xll, xlh, xhl, xhh;

	xll = LoWord(a.lo) + LoWord(b.lo);
	xlh = HiWord(a.lo) + HiWord(b.lo) + HiWord(xll);
	xhl = LoWord(a.hi) + LoWord(b.hi) + HiWord(xlh);
	xhh = HiWord(a.hi) + HiWord(b.hi) + HiWord(xhl);
	if (HiWord(xhh) != 0)
		throw "Overflow in quad adding";

	x.lo = ULongFromWord(xll, xlh);
	x.hi = ULongFromWord(xhl, xhh);
}

static inline void Quad_Sub(const Quad a, const Quad b, Quad& x)
{
	if (a < b)
		throw "Underflow in quad subtraction";

	ULONG xll, xlh, xhl, xhh,
	      all = LoWord(a.lo), bll = LoWord(b.lo),
	      alh = HiWord(a.lo), blh = HiWord(b.lo),
	      ahl = LoWord(a.hi), bhl = LoWord(b.hi),
	      ahh = HiWord(a.hi), bhh = HiWord(b.hi);

	if (all < bll) { all |= 0x10000; ++blh; } xll = all - bll;
	if (alh < blh) { alh |= 0x10000; ++bhl; } xlh = alh - blh;
	if (ahl < bhl) { ahl |= 0x10000; ++bhh; } xhl = ahl - bhl;
	xhh = ahh - bhh;

	x.lo = ULongFromWord(xll, xlh);
	x.hi = ULongFromWord(xhl, xhh);
}

static inline void Quad_Mul(const Quad a, const Quad b, Quad& x)
{
	ULONG xll, xlh, xhl, xhh,
	      all = LoWord(a.lo), bll = LoWord(b.lo),
	      alh = HiWord(a.lo), blh = HiWord(b.lo),
	      ahl = LoWord(a.hi), bhl = LoWord(b.hi),
	      ahh = HiWord(a.hi), bhh = HiWord(b.hi),
		  dum1, dum2, dum3;

	xll = all * bll;
	xlh = alh * bll + all * blh                         + HiWord(xll);
	xhl = ahl * bll + alh * blh + all * bhl             + HiWord(xlh);
	xhh = ahh * bll + ahl * blh + alh * bhl + all * bhh + HiWord(xhl);
	dum1 =            ahh * blh + ahl * bhl + alh * bhh + HiWord(xhh);
	dum2 =                        ahh * bhl + ahl * bhh + HiWord(dum1);
	dum3 =                                    ahh * bhh + HiWord(dum2);
	if (dum1 != 0 || dum2 != 0 || dum3 != 0)
		throw "Overflow in quad multiplying";

	x.lo = ULongFromWord(xll, xlh);
	x.hi = ULongFromWord(xhl, xhh);
}

typedef unsigned short USHORT;

/*
** Modulus must be <= USHRT_MAX, so simple WORDs won't work here
**
** Use a modulus of the form 2^m*16^n (m and n in [0..3])
** to gain the advantage of replacing division by shifting
** and modulus calculation by &-ing
*/
static const int Modulus = 0x8000; // maximum possible
static const int ModulusBits = 3 + 3*4; // == m + n*4
// Number of USHORTs needed to represent a Quad
static const int NumUShort = (2*ulong_bits + (ModulusBits-1)) / ModulusBits;

static inline void Quad_Div(const Quad a, const Quad b, Quad& x, Quad& rem)
{
	if (b == Quad(0))
		throw "Quad division by zero";

	if (a < b)
	{
		x   = 0;
		rem = a;
		return;
	}

	/* Essentially BigNumDiv() from the C Snippets
	**
	** BigNum is Cliff Rhodes' package of unsigned integer math
	** for arbitrarily large numbers,
	** Released to the public domain by Clifford Rhodes, June 15, 1995, with
	** no guarantees of any sort as to accuracy or fitness of use.
	*/
	Quad aa = a, bb = b;
	USHORT at[NumUShort+1], bt[NumUShort],
	       xt[NumUShort], rt[NumUShort],
	       carry, quo;
	long   m1, m2, m3;
	int    alen = NumUShort, blen = NumUShort;
	int    i, j, d, pos;

	memset(&xt[0], 0, NumUShort*sizeof(USHORT));
	memset(&rt[0], 0, NumUShort*sizeof(USHORT));

	for (i = NumUShort-1; i >= 0; --i)
	{
		at[i] = (USHORT)(aa.lo & (ULONG)(Modulus-1)); Quad_Shr(aa, ModulusBits, aa);
		bt[i] = (USHORT)(bb.lo & (ULONG)(Modulus-1)); Quad_Shr(bb, ModulusBits, bb);
	}

	// Assure at[0] == 0, at[1] != 0
	for (pos = 0; pos < alen && at[pos] == 0; ++pos)
		;
	alen -= pos;
	memcpy( &at[1], &at[pos], alen*sizeof(USHORT) );
	at[0] = 0; ++alen;

	// Assure bt[0] != 0
	for (pos = 0; pos < blen && bt[pos] == 0; ++pos)
		;
	blen -= pos;
	memcpy( &bt[0], &bt[pos], blen*sizeof(USHORT) );

	d = Modulus / (bt[0] + 1);
	for (carry = 0, j = alen - 1; j >= 0; --j)
	{
		m1 = (long)d * (long)at[j] + (long)carry;
		at[j] = (USHORT)(m1 & (long)(Modulus-1));
		carry = (USHORT)(m1 >> ModulusBits);
	}
	for (carry = 0, j = blen - 1; j >= 0; --j)
	{
		m1 = (long)d * (long)bt[j] + (long)carry;
		bt[j] = (USHORT)(m1 & (long)(Modulus-1));
		carry = (USHORT)(m1 >> ModulusBits);
	}

	for (j = 0; j < alen - blen; ++j)
	{
		m3 = ((long)at[j] << ModulusBits) | (long)at[j+1];
		if (bt[0] == at[j])
			quo = Modulus - 1;
		else
		{
			m1 = m3 / (long)bt[0];
			quo = (USHORT)m1;
		}
		do
		{
			m1 = (blen > 1) ? ((long)quo * (long)bt[1]) : 0L;
			m2 = ((m3 - ((long)quo * (long)bt[0])) << ModulusBits) + (long)at[j+2];
			if (m1 > m2)
				--quo;
		} while (m1 > m2);

		m2 = 0L;
		for (i = j + blen, pos = blen - 1; i >= j; --i, --pos)
		{
			m1 = (pos >= 0) ? ((long)quo * (long)bt[pos]) : 0L;
			m3 = (long)at[i] - m1 + m2;
			if (m3 < 0L)
			{
				m2 = m3 / (long)Modulus;
				m3 %= (long)Modulus;
				if (m3 < 0L)
				{
					m3 += (long)Modulus;
					--m2;
				}
			}
			else
				m2 = 0L;
			at[i] = (USHORT)m3;
		}
		if (m2 == 0L)
			xt[j] = quo;
		else
		{
			xt[j] = quo - 1;
			carry = 0;
			for (i = j + blen, pos = blen - 1; i >= j; --i, --pos)
			{
				long tmp;

				if (pos >= 0)
					carry += bt[pos];
				tmp = carry + (long)at[i];
				if (tmp >= (long)Modulus)
				{
					tmp -= (long)Modulus;
					carry = 1;
				}
				else
					carry = 0;
				at[i] = (USHORT)tmp;
			}
		}
	}

	carry = 0;
	for (j = alen - blen, pos = 0; j < alen; ++j, ++pos)
	{
		m3 = (long)carry << ModulusBits | (long)at[j];
		rt[pos] = (USHORT)(m3 / d);
		carry   = (USHORT)(m3 % d);
	}

	x = rem = Quad(0);
	for (j = 0; j < alen - blen; ++j)
		{ Quad_Shl(x, ModulusBits, x); x.lo |= xt[j]; }
	for (j = 0; j < blen; ++j)
		{ Quad_Shl(rem, ModulusBits, rem); rem.lo |= rt[j]; }
}

/////////////////////////////////////////////////////////////////////////////
// Quad construction/destruction

// defined inline in Quad.h

/////////////////////////////////////////////////////////////////////////////
// Quad operations

// partially defined inline in Quad.h

// Do not inline the following!
// All Quad_xxx routines are being inlined into these!

Quad Quad::operator+(const Quad q) const
{
	Quad res;
	Quad_Add(*this, q, res);
	return Quad(res);
}

Quad Quad::operator-(const Quad q) const
{
	Quad res;
	Quad_Sub(*this, q, res);
	return Quad(res);
}

Quad Quad::operator*(const Quad q) const
{
	Quad res;
	Quad_Mul(*this, q, res);
	return Quad(res);
}

Quad Quad::operator*(const ULONG l) const
{
	Quad res;
	Quad_Mul(*this,	Quad(l), res);
	return Quad(res);
}

Quad Quad::operator/(const Quad q) const
{
	Quad res, rem;
	Quad_Div(*this, q, res, rem);
	return Quad(res);
}

Quad Quad::operator/(const ULONG l) const
{
	Quad res, rem;
	Quad_Div(*this, Quad(l), res, rem);
	return Quad(res);
}

Quad Quad::operator%(const Quad q) const
{
	Quad res, rem;
	Quad_Div(*this, q, res, rem);
	return Quad(rem);
}

ULONG Quad::operator%(const ULONG l) const
{
	Quad res, rem;
	Quad_Div(*this, Quad(l), res, rem);
	return rem.lo;
}

Quad& Quad::operator+=(const Quad q)
{
	Quad_Add(*this, q, *this);
	return *this;
}

Quad& Quad::operator-=(const Quad q)
{
	Quad_Sub(*this, q, *this);
	return *this;
}

Quad& Quad::operator*=(const Quad q)
{
	Quad_Mul(*this, q, *this);
	return *this;
}

Quad& Quad::operator*=(const ULONG l)
{
	Quad_Mul(*this, Quad(l), *this);
	return *this;
}

Quad& Quad::operator/=(const Quad q)
{
	Quad rem;
	Quad_Div(*this, q, *this, rem);
	return *this;
}

Quad& Quad::operator/=(const ULONG l)
{
	Quad rem;
	Quad_Div(*this, Quad(l), *this, rem);
	return *this;
}

Quad& Quad::operator%=(const Quad q)
{
	Quad res;
	Quad_Div(*this, q, res, *this);
	return *this;
}

Quad& Quad::operator%=(const ULONG l)
{
	Quad res, rem;
	Quad_Div(*this, Quad(l), res, *this);
	return *this;
}

void Quad::FromTime(
	const ULONG day,
	const ULONG month,  // set to 0 for relative time
	const ULONG year,   // set to 0 for relative time
	const ULONG hour,
	const ULONG minute,
	const ULONG second,
	const ULONG millisecond
)
{
	lo = (month == 0 && year == 0) ? day : dmy2jdn(day, month, year);
	hi = 0;
	Quad_Mul(*this,          24, *this);	// hours per day
	Quad_Add(*this, hour       , *this);
	Quad_Mul(*this,          60, *this);	// minutes per hour
	Quad_Add(*this, minute     , *this);
	Quad_Mul(*this,          60, *this);	// seconds per minute
	Quad_Add(*this, second     , *this);
	Quad_Mul(*this,        1000, *this);	// seconds per millisecond
	Quad_Add(*this, millisecond, *this);
	Quad_Mul(*this,       10000, *this);	// milliseconds per 100nanoseconds
}

void Quad::ToAbsTime(
	ULONG& day,
	ULONG& month,
	ULONG& year,
	ULONG& hour,
	ULONG& minute,
	ULONG& second,
	ULONG& millisecond
) const
{
	Quad q(*this), r;

	Quad_Div(q, Quad(10000), q, r);
	Quad_Div(q, Quad( 1000), q, r); millisecond = r.lo;
	Quad_Div(q, Quad(   60), q, r); second      = r.lo;
	Quad_Div(q, Quad(   60), q, r); minute      = r.lo;
	Quad_Div(q, Quad(   24), q, r); hour        = r.lo;
	if (q.hi != 0)
		throw "Quad::ToAbsTime: No valid absolute time";
	jdn2dmy(q.lo, (LONG&)day, (LONG&)month, (LONG&)year);
}

void Quad::ToRelTime(
	ULONG& days,
	ULONG& hours,
	ULONG& minutes,
	ULONG& seconds,
	ULONG& milliseconds) const
{
	Quad q(*this), r;

	Quad_Div(q, Quad(10000), q, r);
	Quad_Div(q, Quad( 1000), q, r); milliseconds = r.lo;
	Quad_Div(q, Quad(   60), q, r); seconds      = r.lo;
	Quad_Div(q, Quad(   60), q, r); minutes      = r.lo;
	Quad_Div(q, Quad(   24), q, r); hours        = r.lo;
	if (q.hi != 0)
		throw "Quad::ToRelTime: No valid relative time";
	days = q.lo;
}

ostream& operator<<(ostream& os, const Quad q)
{
	/*
	** A string of length QLEN must provide enough room for
	**   - octal representation + leading "0"
	**   - decimal representation
	**   - hexadecimal representation + leading "0x"
	*/
	const int QLEN = 30;

	Quad qq = q, q0(0), c;
	char *qs;
	int pos, len = 0, prefixLen;
	ULONG base;
	long iosFlags;
	char fill;
	int width;

	width = os.width();

	pos = (QLEN > width) ? QLEN : width;
	qs = new char[pos+1];

	iosFlags = os.flags();

	     if (iosFlags & ios::oct) { base = 8;  prefixLen = 1; }
	else if (iosFlags & ios::hex) { base = 16; prefixLen = 2; }
	else /* if (iosFlags & ios::dec) */ { base = 10; prefixLen = 0; }

	qs[pos] = '\0';
	do
	{
		Quad_Div(qq, Quad(base), qq, c);

		if (base == 16 && c.lo > 9)
			qs[--pos] = ((iosFlags & ios::uppercase) ? 'A' : 'a') + ((int)c.lo - 10);
		else
			qs[--pos] = '0' + (int)c.lo;
	} while (qq != q0);

	if (iosFlags & ios::showbase)
		switch (base)
		{
			case 16 : qs[--pos] = 'x'; /* FALLTHROUGH */
			case 8  : qs[--pos] = '0'; /* FALLTHROUGH */
			case 10 : break;
		}

	len = QLEN - pos;
	if (width != 0 && width > len)
	{
		fill = os.fill();
		if (iosFlags & ios::left)
		{
			// base prefix, then number, then fill characters
			memcpy(&qs[0], &qs[pos], len);
			pos = 0;
			memset(&qs[len], fill, width - len);
			qs[width] = '\0';
		}
		else
		if ( (iosFlags & ios::right) ||
		    ((iosFlags & ios::internal) && !(iosFlags & ios::showbase)))
		{
			// fill characters, then base prefix, then number
			pos -= width - len;
			memset(&qs[pos], fill, width - len);
		}
		else
			if (iosFlags & ios::internal) // && (iosFlags & ios::showbase
		{
			// base prefix, then fill characters, then number
			memcpy(&qs[pos-(width-len)], &qs[pos], prefixLen);
			memset(&qs[pos-(width-len)+prefixLen], fill, width-len);
			pos -= width-len;
		}
	}

	os << &qs[pos];

	delete[] qs;

	return os;
}

////////////////////////////////////////////////////////////////////////////
